/* --------------------------------------------------------------------------
        DNSInfo.cpp

        Copyright (C) 2004 Jason Weiler, All Rights Reserved
   -------------------------------------------------------------------------- */

#include <windows.h>
#include <stdio.h>

#include <string>

#include "DNSInfoConstants.h"
#include "DNSInfoTypes.h"
#include "DNSInfoConfig.h"
#include "DNSPacketParsing.h"

#include "DNSServerList.h"
#include "DataBuffer.h"
#include "MacroUtils.h"

#include "IPTransactionNetworkFactory.h"

#include "CAssert.h"
CASSERT(sizeof(struct DNSPacketHeader_t) == DNSPacketHeaderSize);
CASSERT(sizeof(struct DNSQuestion_t) == DNSQuestionSize);
CASSERT(sizeof(struct DNSResource_t) == DNSResourceSize+2); // we know that this structure pads out a little. That's ok though.
CASSERT(sizeof(struct DNSSOAResource_t) == DNSSOAResourceSize);


namespace
{
	
long EncodeDNSName(const std::string & sName, CDataBuffer & outBuf)
{
	bool bDone = false;

	unsigned char ucOffset = 0;

	std::string::size_type findPos;
	std::string::size_type searchPos = 0;

	do
	{
		findPos = sName.find_first_of(".", searchPos);
		
		if (findPos == std::string::npos)
		{
			findPos = sName.size();
			bDone = true;
		}

		// write the length of this segment
		ucOffset = (unsigned char)(findPos - searchPos);
		outBuf.Append(&ucOffset, sizeof(ucOffset));

		// copy the contents of this segment
		outBuf.Append((const unsigned char*)sName.c_str()+searchPos, findPos - searchPos);
		searchPos = findPos+1;
	}
	while (!bDone);

	// zero-length segment at the end
	ucOffset = 0;
	outBuf.Append(&ucOffset, sizeof(ucOffset));

	return outBuf.Size();
}

/* return the number of bytes in the inData that were freshly parsed. Bytes
 *	that are part of the DNS name-compression scheme don't count
 * pReadPtr - a ptr to the next byte to be interpretted
 * pInBuf - a pointer to the top of entire block of DNS data. We need this for
 *			compressed DNS names.
 */
long DecodeDNSName(const unsigned char * pReadPtr, const char * pInBuf, std::string & sOutStr)
{
	const char * inPtr = pInBuf;
	long bufferBytesUsed = 0;

	bool bCompressed = false;

	sOutStr.clear();

	while (0 != *inPtr)
	{
		unsigned short tmpLen = ntohs(*(unsigned short*)inPtr);
		if ( (tmpLen & DNS_NAME_COMPRESSION_FLAG) == DNS_NAME_COMPRESSION_FLAG)
		{
			// we're using DNS name compression.
			unsigned short tmpOffset = tmpLen & NOT_DNS_NAME_COMPRESSION_MASK;

			inPtr = (const char*)pReadPtr + tmpOffset;
			if (bCompressed == false)
			{
				// Don't count these if we're already compressing
				bufferBytesUsed += 2;
			}
			bCompressed = true;

			/* we have to continue here because there might be
				multiple compression indirections */
			continue;
		}

		sOutStr.append(inPtr+1, *inPtr);

		if (bCompressed == false)
		{
			// only record the byte count if these are NEW bytes being used!!
			bufferBytesUsed += *inPtr+1;
		}

		inPtr += *inPtr + 1;
		if (*inPtr != 0)
		{
			sOutStr += ".";
		}
	}

	if (!bCompressed)
	{
		// only record the byte count if these are NEW bytes being used!!
		// count the zero-length segment at the end
		++bufferBytesUsed;
	}


	if (0 == sOutStr.size())
	{
		sOutStr = "{empty}";
	}

	return bufferBytesUsed;
}

long MakeDNSRequest(DNSRequest_t * req, CDataBuffer & replyBuf)
{
	// configure the main request header
	DNSPacketHeader_t packHdr;

	packHdr.Identification = 100;
	packHdr.usFlags = 0;
	packHdr.usNumQuestions = 1;
	packHdr.usNumAnswerRRs = 0;
	packHdr.usNumAuthorityRRs = 0;
	packHdr.usNumAdditionalRRs = 0;

	// Do we want the server to do our recursion for us?
	if (true == req->bUseRecursion)
		FLAGS_SetRD(packHdr.usFlags, 1);
	else
		FLAGS_SetRD(packHdr.usFlags, 0);

	FLAGS_SetOpcode(packHdr.usFlags, req->ucOpcode);

	// put all the Header fields in network byte order
	packHdr.Identification = htons(packHdr.Identification);
	packHdr.usFlags = htons(packHdr.usFlags);
	packHdr.usNumQuestions = htons(packHdr.usNumQuestions);
	packHdr.usNumAnswerRRs = htons(packHdr.usNumAnswerRRs);
	packHdr.usNumAuthorityRRs = htons(packHdr.usNumAuthorityRRs);
	packHdr.usNumAdditionalRRs = htons(packHdr.usNumAdditionalRRs);

	// Encode the query name.
	CDataBuffer encName;
	long encNameLen = EncodeDNSName(req->sAddress, encName);

	// Put the rest of the Query in network byte order
	DNSQuestion_t qData;
	qData.usType = htons(req->usType);
	qData.usClass = htons(req->usClass);


	// Finally, concatenate all the query-parts together for sending
	CDataBuffer sendBuf;
	sendBuf.Clear();

	sendBuf.Append((const unsigned char*)&packHdr, DNSPacketHeaderSize);
	sendBuf.Append(encName.GetDataPtr(NULL), encNameLen);
	sendBuf.Append((const unsigned char*)&qData, DNSQuestionSize);

	// Create our network object
	IPTransactionNetwork * tNet = CreateIPTransactionNetwork(IPTRANSACTIONNETWORK_UDP);

	// Start and configure the object
	if (0 != tNet->Startup())
	{
		// error
		DestroyIPTransactionNetwork(tNet);
		tNet = NULL;
		return -1;
	}

	tNet->SetServerPort(req->ulServer, DNS_DEFAULT_PORT);

	if (false == tNet->Connect())
	{
		// error
		tNet->Shutdown();
		DestroyIPTransactionNetwork(tNet);
		return -2;
	}
	
	if (false == tNet->SendTransaction(sendBuf, replyBuf))
	{
		// error
		tNet->Disconnect();
		tNet->Shutdown();
		DestroyIPTransactionNetwork(tNet);
		tNet = NULL;
		
		return -3;
	}

	tNet->Disconnect();
	tNet->Shutdown();
	DestroyIPTransactionNetwork(tNet);
	tNet = NULL;

	return 0;
}


long CreateResourceInfoOutput(const unsigned char * rawData, const unsigned char * rsrcPtr,
							  const DNSResource_t & rsrcHdr, std::string & sOutStr)
{
	// The Resource Info differs for each type of Resource. This function merely
	//  parses each type of data and puts an appropriate bit of output in the
	//  outStr parameter
	switch(rsrcHdr.usType)
	{
		case QUERY_TYPE_A: // 1
		{
			char tmpStr[64];
			wsprintf(tmpStr, "IP:%lu.%lu.%lu.%lu", *(rsrcPtr), *(rsrcPtr+1),
															   *(rsrcPtr+2), *(rsrcPtr+3));
			sOutStr += tmpStr;
		}
		break;

		case QUERY_TYPE_NS: // 2
			sOutStr += "Name Svr: ";
			rsrcPtr += DecodeDNSName(rawData, (const char*)rsrcPtr, sOutStr);
		break;

		case QUERY_TYPE_MD:	// 3
			sOutStr += "MADNAME: ";
			rsrcPtr += DecodeDNSName(rawData, (const char*)rsrcPtr, sOutStr);
			sOutStr += "OBSOLETE";
		break;

		case QUERY_TYPE_MF:	// 4
			sOutStr += "MADNAME: ";
			rsrcPtr += DecodeDNSName(rawData, (const char*)rsrcPtr, sOutStr);
			sOutStr += "OBSOLETE";
		break;

		case QUERY_TYPE_CNAME: // 5
			sOutStr += "Canonical Name: ";
			rsrcPtr += DecodeDNSName(rawData, (const char*)rsrcPtr, sOutStr);
		break;

		case QUERY_TYPE_SOA:	// 6
		{
			DNSSOAResource_t soaRec;
			sOutStr += "Name Svr: ";
			rsrcPtr += DecodeDNSName(rawData, (const char*)rsrcPtr, sOutStr);
			sOutStr += "\n";

			sOutStr.append("\tReponsible Mailbox: ");

			rsrcPtr += DecodeDNSName(rawData, (const char*)rsrcPtr, sOutStr);

			rsrcPtr += ParseSOAResourceData(rsrcPtr, soaRec);

			char tmpStr[64];

			wsprintf(tmpStr, "\n\tSerial:  %lu", soaRec.ulSerial);
			sOutStr += tmpStr;
			wsprintf(tmpStr, "\n\tRefresh: %lu", soaRec.ulRefresh);
			sOutStr += tmpStr;
			wsprintf(tmpStr, "\n\tRetry:   %lu", soaRec.ulRetry);
			sOutStr += tmpStr;
			wsprintf(tmpStr, "\n\tExpire:  %lu", soaRec.ulExpire);
			sOutStr += tmpStr;
			wsprintf(tmpStr, "\n\tMinimum: %lu", soaRec.ulMinimum);
			sOutStr += tmpStr;
		}
		break;
		case QUERY_TYPE_MB:		// 7
			sOutStr += "Mailbox Host: ";
			rsrcPtr += DecodeDNSName(rawData, (const char*)rsrcPtr, sOutStr);
		break;

		case QUERY_TYPE_MG:		// 8
			sOutStr += "Group Mailbox: ";
			rsrcPtr += DecodeDNSName(rawData, (const char*)rsrcPtr, sOutStr);
		break;

		case QUERY_TYPE_MR:		// 9
			sOutStr += "Mailbox Rename: ";
			rsrcPtr += DecodeDNSName(rawData, (const char*)rsrcPtr, sOutStr);
		break;

		case QUERY_TYPE_WKS:	// 11
		case QUERY_TYPE_TXT:	// 16
			// because the text format is unknown, we are treating it as raw data.
			// fall through
		case QUERY_TYPE_NULL:	// 10
		{
			// this field can be anything, so we're going to print
			//  it as a hex dump in rows of 16 bytes
			long i=0;
			char tmpStr[16];

			sOutStr += "\n\tData = ";

			while (i<rsrcHdr.usRDataLen)
			{
				wsprintf(tmpStr, "%02X ", *rsrcPtr);
				sOutStr += tmpStr;

				++rsrcPtr;
				++i;
	
				// add a carriage return every 16 bytes
				if (i%16 == 0)
				{
					sOutStr += "\n\t       ";
				}

			}
		}
		break;


		case QUERY_TYPE_PTR:	// 12

		sOutStr.append("Ptr: ");

		rsrcPtr += DecodeDNSName(rawData, (const char*)rsrcPtr, sOutStr);
		break;

		case QUERY_TYPE_HINFO:	// 13
		{
			unsigned long len = *rsrcPtr;
			++rsrcPtr;

			sOutStr.append("CPU: ");
			sOutStr.append((const char*)rsrcPtr, len);

			rsrcPtr += len;
			
			len = *rsrcPtr;
			++rsrcPtr;

			sOutStr.append("OS: ");
			sOutStr.append((const char*)rsrcPtr, len);

			rsrcPtr += len;
		}
		break;

		case QUERY_TYPE_MINFO:	// 14
		{
			sOutStr += "MailList Owner: ";
			rsrcPtr += DecodeDNSName(rawData, (const char*)rsrcPtr, sOutStr);
			sOutStr += "MailList Error: ";
			rsrcPtr += DecodeDNSName(rawData, (const char*)rsrcPtr, sOutStr);
		}
		break;

		case QUERY_TYPE_MX:		// 15
		{
			char tmpStr[64];

			unsigned short pref = ntohs(*(unsigned short*)rsrcPtr);
			rsrcPtr += sizeof(pref);
		
			wsprintf(tmpStr, "Mail Exch (pref: %u): ", pref);
			sOutStr += tmpStr;

			rsrcPtr += DecodeDNSName(rawData, (const char*)rsrcPtr, sOutStr);
		}
		break;


		default:
		{
			sOutStr += "\tUnknown Resource Type";
		}
		break;
	}

	return 0;
}

long DumpDNSReply(CDataBuffer & replyBuf)
{

	if (true == replyBuf.Empty())
	{
		// didn't get a reply...error
		return -3;
	}

	// Now that the request has been made and a reply has been received,
	//  we need to parse it for output.
	size_t replyBufLen;
	const unsigned char * parsePtr = replyBuf.GetDataPtr(&replyBufLen);

	// Parse the DNS packet header
	DNSPacketHeader_t dnsHdr;
	
	parsePtr += ParseDNSPacketHeader(parsePtr, dnsHdr);


	std::string sTmpStr;	// generic answer string buffer

	// ---------------------------------------
	// Dump the header info

	// REPLY CODE
	if (FLAGS_GetRCode(dnsHdr.usFlags) != 0)
	{
		printf("Reply Code: %u", FLAGS_GetRCode(dnsHdr.usFlags)); 

		switch (FLAGS_GetRCode(dnsHdr.usFlags))
		{
		case 1:			// Format Error
			printf (" - Unable to Interpret Query\n");
			break;

		case 2:			// Server Failure
			printf (" - Server Failure\n");
			break;

		case 3:			// Name Error
			printf (" - Referenced Domain Name Not Found\n");
			break;

		case 4:			// Function not implemented
			printf (" - Feature Not Supported By Server\n");
			break;

		case 5:			// Request Refused
			printf (" - Request Refused By Server\n");
			break;

		default:
			printf (" - Unknown Error\n");
			break;
		}
	}

	// TRUNCATION
	if (1 == FLAGS_GetTC(dnsHdr.usFlags))
		printf ("Warning: Answer is truncated!\n");



	// AUTHORITATIVENESS
	if (1 == FLAGS_GetAA(dnsHdr.usFlags))
		printf("Authoritative Answer\n");
	else
		printf("Non-Authoritative Answer\n");


	// ---------------------------------------
	// Parse and output the QUESTION Resources
	DNSQuestion_t dnsQstn;
	printf ("Questions:\n");
	for (unsigned long i=0; i<dnsHdr.usNumQuestions; ++i)
	{
		sTmpStr.clear();
		parsePtr += DecodeDNSName(replyBuf.GetDataPtr(NULL), (const char *)parsePtr, sTmpStr);
		parsePtr += ParseDNSQuestion(parsePtr, dnsQstn);

		printf("  {%s,%s}\t\t%s\n", GetQueryTypeStr((EQueryType)dnsQstn.usType),
								  GetQueryClassStr((EQueryClass)dnsQstn.usClass),
								  sTmpStr.c_str());
	}


	
	// -------------------------------------
	// Parse and output the ANSWER Resources
	DNSResource_t dnsRsrc;

	if (dnsHdr.usNumAnswerRRs > 0)
	{
		printf ("Answers:\n");
	}
	for (unsigned long i=0; i<dnsHdr.usNumAnswerRRs; ++i)
	{
		sTmpStr.clear();
		parsePtr += DecodeDNSName(replyBuf.GetDataPtr(NULL), (const char *)parsePtr, sTmpStr);
		parsePtr += ParseDNSResource(parsePtr, dnsRsrc);
		printf("  {%s,%s} TTL: %u\t%s\t", GetQueryTypeStr((EQueryType)dnsRsrc.usType),
								  GetQueryClassStr((EQueryClass)dnsRsrc.usClass),
								  dnsRsrc.ulTTL,
								  sTmpStr.c_str());

		sTmpStr.clear();
		
		CreateResourceInfoOutput(replyBuf.GetDataPtr(NULL), parsePtr, dnsRsrc, sTmpStr);
		parsePtr += dnsRsrc.usRDataLen;
		
		printf("%s\n", sTmpStr.c_str());
	}

	// ----------------------------------------
	// Parse and output the AUTHORITY Resources
	if (dnsHdr.usNumAuthorityRRs > 0)
	{
		printf ("Authority:\n");
	}
	for (unsigned long i=0; i<dnsHdr.usNumAuthorityRRs; ++i)
	{
		sTmpStr.clear();
		parsePtr += DecodeDNSName(replyBuf.GetDataPtr(NULL), (const char *)parsePtr, sTmpStr);
		parsePtr += ParseDNSResource(parsePtr, dnsRsrc);
		printf("  {%s,%s} TTL: %u\t%s\t", GetQueryTypeStr((EQueryType)dnsRsrc.usType),
										  GetQueryClassStr((EQueryClass)dnsRsrc.usClass),
										  dnsRsrc.ulTTL,
										  sTmpStr.c_str());

		sTmpStr.clear();

		CreateResourceInfoOutput(replyBuf.GetDataPtr(NULL), parsePtr, dnsRsrc, sTmpStr);
		parsePtr += dnsRsrc.usRDataLen;

		printf("%s\n", sTmpStr.c_str());
	}


	// -----------------------------------------
	// Parse and output the ADDITIONAL Resources
	if (dnsHdr.usNumAdditionalRRs > 0)
	{
		printf ("Additional:\n");
	}
	for (unsigned long i=0; i<dnsHdr.usNumAdditionalRRs; ++i)
	{
		sTmpStr.clear();
		parsePtr += DecodeDNSName(replyBuf.GetDataPtr(NULL), (const char *)parsePtr, sTmpStr);
		parsePtr += ParseDNSResource(parsePtr, dnsRsrc);
		printf("  {%s,%s} TTL: %u\t%s\t", GetQueryTypeStr((EQueryType)dnsRsrc.usType),
										  GetQueryClassStr((EQueryClass)dnsRsrc.usClass),
										  dnsRsrc.ulTTL,
										  sTmpStr.c_str());

		sTmpStr.clear();
		CreateResourceInfoOutput(replyBuf.GetDataPtr(NULL), parsePtr, dnsRsrc, sTmpStr);
		parsePtr += dnsRsrc.usRDataLen;
		
		printf("%s\n", sTmpStr.c_str());
	}
	
	return 0;
}

void Usage()
{
	fputs("Usage: DNSinfo [-?] [-svr ServerIP] [-type Type] [-recurse|-norecurse}] Address\n\n", stderr);
	fputs("         -?       - This message\n", stderr);
	fputs("         -recurse - Tell the DNS server to use recursive lookup (default)\n", stderr);
	fputs("         -norecurse - Tell the DNS server to use non-recursive lookup\n", stderr);
	fputs("         -svr     - Specify a DNS server that should be used\n", stderr);
	fputs("         -type    - Specify the type of the DNS query that should be made\n\n", stderr);
	fputs("  where: Type       ::= A | NS | MD | MF | CNAME | SOA | MB |\n", stderr);
	fputs("                         MG | MR | NULL | WKS | PTR | HINFO |\n", stderr);
	fputs("                         MINFO | MX | TXT | AXFR | ANY}\n", stderr);
	fputs("         ServerIP  ::= the ip address of a valid DNS server\n", stderr);
	fputs("         Address    ::= the dns name or ip address of an internet server\n\n", stderr);

	return;
}


} // anon namespace


int main(int argc, const char *argv[])
{
	DNSInfoConfig Config;
	// Handle arguments
	
	if (false == Config.ParseCommandLine(argc, argv))
	{
		Usage();
		return 0;
	}

	// Set up our request structure with the settings specified by the user.
	DNSRequest_t dnsReq;
	
	dnsReq.bUseRecursion = Config.GetUseRecursion();
	dnsReq.usType = (unsigned short)Config.GetQueryType();
	dnsReq.usClass = (unsigned char)Config.GetQueryClass();		// always internet for now
	dnsReq.ucOpcode = (unsigned char)Config.GetQueryOpcode();
	dnsReq.sAddress = Config.GetQueryAddress();



	DNSServerList dnsSvrList;
	dnsSvrList = Config.GetDNSServerList();

	// scan the registry for DNS servers. If the user specified their own server,
	//  it will get first crack at our request. If that fails, these servers will
	//  get a chance at it.
	dnsSvrList.ScanRegistry();
	if (0 >= dnsSvrList.GetServerCount())
	{
		// there are no servers to speak of
		printf ("Unable to locate any DNS servers!\n");
		return -1;
	}

	CDataBuffer replyBuf;	

	if (0 == dnsSvrList.GetServerCount())
	{
		printf ("No DNS servers to query!\n");
	}

	for (unsigned int iSvrIdx=0; iSvrIdx < dnsSvrList.GetServerCount(); ++iSvrIdx)
	{
		dnsReq.ulServer = dnsSvrList.GetNthAddr(iSvrIdx);
		if (INADDR_NONE == dnsReq.ulServer)
		{
			// skip any non-starters
			continue;
		}

		// a little feedback about where we're going
		printf("Querying %lu.%lu.%lu.%lu\n", dnsReq.ulServer & 0x000000FF
									, (dnsReq.ulServer >> 8) & 0xFF
									, (dnsReq.ulServer >> 16) & 0xFF
									, (dnsReq.ulServer >> 24) & 0xFF );

		// Make the request and get the raw answer buffer
		if (0 == MakeDNSRequest(&dnsReq, replyBuf))
		{
			// Finally, dump the contents of the reply
			if (0 <= DumpDNSReply(replyBuf))
			{
				// success
				break;
			}
			else
			{
				// parsing error...this should be rare indeed
				printf ("Error Parsing Reply from DNS Server\n");
			}
		}
		else
		{
			// error in the request
			printf ("Error Sending Request to DNS Server\n");
		}

	}

	return 0;
}

