/* --------------------------------------------------------------------------
        DNSServerList.cpp

        Copyright (C) 2004 Jason Weiler, All Rights Reserved
   -------------------------------------------------------------------------- */

#include <windows.h>
#include <algorithm>
#include <string>
#include <vector>

#include "DNSServerList.h"

const unsigned long RegQueryBufSize = (1024);

unsigned long DNSServerList::ScanRegistry()
{
	unsigned long ulServerCount = 0;
	// The registry key for looking up DNS servers varies with the operating 
	// system.  Test version first.
	OSVERSIONINFO verinfo = {sizeof(verinfo)};
	GetVersionEx(&verinfo);

	if ( VER_PLATFORM_WIN32_WINDOWS == verinfo.dwPlatformId )
	{
		// Trying Win9X NameServer
		ulServerCount += EnumDNSServersList("System\\CurrentControlSet\\Services\\VxD\\MSTCP", "NameServer");
		// Trying Win9X DhcpNameServer
		ulServerCount += EnumDNSServersList("System\\CurrentControlSet\\Services\\VxD\\MSTCP", "DhcpNameServer");
		// Trying Win9X OptionInfo
		ulServerCount += Alt_EnumDNSServers_9X("System\\CurrentControlSet\\Services\\VxD\\DHCP", "OptionInfo");
	}
	else if ( VER_PLATFORM_WIN32_NT == verinfo.dwPlatformId )
	{
		// Trying WinNT NameServer
		ulServerCount += EnumDNSServersList("System\\CurrentControlSet\\Services\\TCPIP\\Parameters", "NameServer");
		// Trying WinNT DhcpNameServer
		ulServerCount += EnumDNSServersList("System\\CurrentControlSet\\Services\\TCPIP\\Parameters", "DhcpNameServer");
		if (verinfo.dwMajorVersion == 5)		// Win2000 only
		{
			// Trying Win2000 Interfaces\\*\\NameServer
			ulServerCount += EnumDNSServers_2000("System\\CurrentControlSet\\Services\\Tcpip\\Parameters\\Interfaces", "NameServer");
			// Trying Win2000 Interfaces\\*\\DhcpNameServer
			ulServerCount += EnumDNSServers_2000("System\\CurrentControlSet\\Services\\Tcpip\\Parameters\\Interfaces", "DhcpNameServer");
		}

	}

	return ulServerCount;
}	

bool DNSServerList::AddServer(unsigned long ulAddr)
{
	if (m_ServerAddrs.end() == std::find(m_ServerAddrs.begin(), m_ServerAddrs.end(), ulAddr))
	{
		// if the server isn't already in the list, then add it.
		m_ServerAddrs.push_back(ulAddr);

		return true;
	}
	
	return false;
}

unsigned long DNSServerList::EnumDNSServersList(const std::string & sRegKey, const std::string & sRegName)
{
	unsigned long ulServerCount = 0;

	// Ask for a space-separated list of DNS servers
	unsigned char RegBuffer[RegQueryBufSize];
	RegBuffer[0] = '\0';
	unsigned long ulRegBufferLen = sizeof(RegBuffer);
	

	if (true == QueryRegistryKey( HKEY_LOCAL_MACHINE, sRegKey.c_str(), sRegName.c_str(),
									(BYTE*)RegBuffer, &ulRegBufferLen, NULL ))
	{
		// Parse through the servers and put them into our list
		ulServerCount = ParseAddDNSServerList((char*)RegBuffer);
	}

	return ulServerCount;
}

unsigned long DNSServerList::EnumDNSServers_2000(const std::string & sRegKey,
													  const std::string & sRegName)
{
	// Ask for a space-separated list of DNS servers
	unsigned long ulServerCount = 0;

	unsigned char RegBuffer[RegQueryBufSize];
	RegBuffer[0] = '\0';
	unsigned long ulRegBufferLen = sizeof(RegBuffer);
	FILETIME ftTime;

	HKEY hBaseKey = 0;
	unsigned long ulKeyIdx = 0;

	if (ERROR_SUCCESS != RegOpenKeyEx(HKEY_LOCAL_MACHINE, sRegKey.c_str(), 0, KEY_READ, &hBaseKey))
	{
		return 0;
	}


	char szKeyName[MAX_PATH];
	DWORD dwKeyNameLen = MAX_PATH;

	while (ERROR_SUCCESS == RegEnumKeyEx(hBaseKey, ulKeyIdx, szKeyName, &dwKeyNameLen, NULL, NULL, NULL, &ftTime))
	{
		RegBuffer[0] = '\0';
		ulRegBufferLen = sizeof(RegBuffer);
		QueryRegistryKey(hBaseKey, szKeyName, sRegName.c_str(), (BYTE*)RegBuffer, &ulRegBufferLen, NULL );

		// Parse through the servers and put them into our list
		ulServerCount += ParseAddDNSServerList((char*)RegBuffer);

		++ulKeyIdx;
		dwKeyNameLen = MAX_PATH;
		
	}

	RegCloseKey(hBaseKey);
	hBaseKey = NULL;

	return ulServerCount;
}

unsigned long DNSServerList::Alt_EnumDNSServers_9X(const std::string & sRegKey,
														const std::string & sRegName)
{
	// Ask for a space-separated list of DNS servers
	unsigned long ulServerCount = 0;

	unsigned char RegBuffer[RegQueryBufSize];
	RegBuffer[0] = '\0';
	unsigned long ulRegBufferLen = sizeof(RegBuffer);
	FILETIME ftTime;

	HKEY hBaseKey = 0;
	unsigned long ulKeyIdx = 0;

	if (ERROR_SUCCESS != RegOpenKeyEx(HKEY_LOCAL_MACHINE, sRegKey.c_str(), 0, KEY_READ, &hBaseKey))
	{
		return 0;
	}


	char keyName[MAX_PATH];
	DWORD keyNameLen = MAX_PATH;

	while (ERROR_SUCCESS == RegEnumKeyEx(hBaseKey, ulKeyIdx, keyName, &keyNameLen, NULL, NULL, NULL, &ftTime))
	{
		RegBuffer[0] = '\0';
		ulRegBufferLen = sizeof(RegBuffer);
		QueryRegistryKey(hBaseKey, keyName, sRegName.c_str(), (BYTE*)RegBuffer, &ulRegBufferLen, NULL );
		unsigned char * sectPtr = (unsigned char*)RegBuffer;
		unsigned char * parsePtr;
		unsigned char type = 0;
		unsigned char len = 0;
		unsigned char count = 0;
		long i = 0;

		// The sectPtr ALWAYS points to the start of the current data
		// The parsePtr moves through the data
		while (((RegBuffer + ulRegBufferLen) > sectPtr) && (0xFF != *sectPtr) )
		{
			type = *sectPtr;
			len = *(sectPtr+1);
			sectPtr+=2;
			

			if (0x06 == type)
			{
				// this is the DNS list...so parse it.
				parsePtr = sectPtr;
				count = len>>2;
				for (i=0; i<count; i++)
				{
					
					if (0 == AddServer(*(unsigned long*)parsePtr))
					{
						++ulServerCount;
					}

					parsePtr+=sizeof(unsigned long);
				}
			}

			// finally, move on to the next section.
			sectPtr += len;
		}
		
		++ulKeyIdx;
		keyNameLen = MAX_PATH;
		
	}

	RegCloseKey(hBaseKey);
	hBaseKey = NULL;

	return ulServerCount;
}

// Parse a list of servers that are comma- or space-separated
unsigned long DNSServerList::ParseAddDNSServerList(const std::string & sList)
{
	unsigned long ulServerCount = 0;

	std::string::size_type spos = sList.find_first_not_of(", \t\n\r", 0);
	std::string::size_type epos = sList.find_first_of(", \t\n\r", spos);
	
	if (std::string::npos == epos) epos = sList.size();

	while (std::string::npos != spos)
	{
		// make sure the name-server IP is really an IP
		unsigned long ulTmpServer = inet_addr(sList.substr(spos, epos-spos).c_str());
		if (INADDR_NONE != ulTmpServer)
		{
			if (true == AddServer(ulTmpServer))
			{
				++ulServerCount;
			}
		}

		spos = sList.find_first_not_of(", \t\n\r", epos);
		epos = sList.find_first_of(", \t\n\r", spos);
		
		if (std::string::npos == epos)
			epos = sList.size();
	}

	return ulServerCount;
}

bool DNSServerList::QueryRegistryKey(HKEY hBaseKey, const std::string & sKeyPath,
									  const std::string & sKeyName,
									  unsigned char * outValue,
									  unsigned long * outValueLen,
									  unsigned long * outValueType)
{
	HKEY hKeyResult = 0;
	LONG lReturn = 0;

	DWORD dwType;
	

	// handle of open key                             
	lReturn = RegOpenKeyEx( hBaseKey,
							sKeyPath.c_str(),// address of name of subkey to open	
							0,		//reserved 
							KEY_ALL_ACCESS,		// security access mask 
							&hKeyResult );		//address of handle of open key 

	if (ERROR_SUCCESS != lReturn)
	{
		return false;
	}

	lReturn = RegQueryValueEx(hKeyResult,
							  sKeyName.c_str(),
							  0,
							  &dwType,
							  outValue,  // returned data from the Query
							  outValueLen);

	// We no longer need the HKEY, so release it
	RegCloseKey(hKeyResult);
	hKeyResult = NULL;;
	

	if (ERROR_SUCCESS != lReturn )
	{
		return false;
	}

	if (NULL != outValueType)
	{
		*outValueType = dwType;
	}

	// success
	return true;
}


